"""
Extends Redis with DataStructure
"""

from redis import Redis
import redis_collections

class RedisCollection(Redis):

    def List(self, key):
        return redis_collections.List(key=key, redis=self)

    def Dict(self, key):
        return redis_collections.Dict(key=key, redis=self)

    def Set(self, key):
        return redis_collections.Set(key=key, redis=self)



from elasticsearch import Elasticsearch
from elasticsearch.helpers import scan as es_scan, bulk as es_bulk
import elasticsearch_dsl as dsl

# ElasticSearch
class ElasticSearchIndex(object):
    def __init__(self, url):
        uri, index = url.split("/", 2)
        self._uri = uri
        self.client = Elasticsearch(uri)
        self.index_name = index

    def use(self, doc_type):
        return Doc(client=self.client,
                                index_name=self.index_name,
                                doc_type=doc_type)

    def delete(self):
        self.client.indices.delete(index=self.index_name, ignore=[400, 404])

class Doc(object):

    def __init__(self, client, index_name, doc_type):
        self.client = client
        self.index_name = index_name
        self.doc_type = doc_type

    def index(self, id, doc, upsert=True, ignore=[404], **kwargs):
        """
        To create or update doc
        :param id: int
        :param doc: dict
        :param upsert: bool
        :param ignore: dict
        :param kwargs:
        :return:
        """
        body = doc
        if upsert:
            body = {
                "doc": doc,
                "doc_as_upsert": True
            }
        return self.client.update(index=self.index_name,
                                  doc_type=self.doc_type,
                                  id=id,
                                  body=body,
                                  ignore=ignore,
                                  **kwargs)

    def delete(self, id, ignore=[404], **kwargs):
        return self.client.delete(index=self.index_name,
                                  doc_type=self.doc_type,
                                  id=id,
                                  ignore=ignore
                                  **kwargs)

    def get(self, id, **kwargs):
        return self.client.get(index=self.index_name,
                               doc_type=self.doc_type,
                               id=id,
                               **kwargs)

    def create(self, id, doc, **kwargs):
        """
        To create a new index. If it exists, will throw error
        :param id: int
        :param doc: dict
        :param kwargs:
        :return:
        """
        return self.client.create(index=self.index_name,
                                  doc_type=self.doc_type,
                                  id=id,
                                  body=doc,
                                  **kwargs)


    def search(self, *args, **kwargs):
        return dsl.Search(using=self.client,
                          index=self.index_name,
                          doc_type=self.doc_type)

    def delete_all_documents(self):
        scroll = '5m'
        docs = es_scan(self.client,
                       query={},
                       index=self.index_name,
                       doc_type=self.doc_type)
        for d in docs:
            self.delete(id=d["_id"])



""" __about__ """

__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "magic-xxl"
__version__ = "0.6.0"
__summary__ = "A collection librairies to work with Flask-Magic"
__uri__ = "https://github.com/mardix/magic-xxl"
__author__ = "Mardix"
__email__ = "mardix@pylot.io"
__license__ = "MIT"
__copyright__ = "(c) 2016 %s" % __author__



"""
APM is app performance monitoring service.
Currently uses NewRelic, to monitor the app

How to use

from magic_xxl.flask import newrelic_apm as apm

apm.apm.set_name("www")

Requirements:
    config:
        APPLICATION_NAME
        NEWRELIC_LICENSE_KEY
"""

from flask_magic import init_app, get_env
import newrelic.agent

newrelic.agent.initialize()
settings = newrelic.agent.global_settings()

class NewRelicAPM(object):
    _app = None
    _app_name = "www"

    def init_app(self, app):

        wp_application_name = app.config.get("APPLICATION_NAME")
        env_name = get_env().lower()
        app_name = "%s/%s:%s" % (wp_application_name, self._app_name, env_name)

        settings.app_name = app_name
        settings.license_key = app.config.get("NEWRELIC_LICENSE_KEY")

    def set_name(self, app_name):
        self._app_name = app_name

apm = NewRelicAPM()
init_app(apm.init_app)




from redis_collection import RedisCollection
from taskman import TaskMan
from elasticsearch_index import ElasticSearchIndex

__all__ = [
    "RedisCollection",
    "TaskMan",
    "ElasticSearchIndex"
]


"""
TaskMan a message queue manager for SQS that executes jobs in the background

It put data in 2 queues:
    - queued
    - finished
    
queued: messages contain runnable messages, ie: function to run for task
finished: will contained the finished results which will contain completed
            or failed data


mytaskman = TaskMan("myQname", "***", "***")

- To add a task(job)
mytaskman.add_task(my_function_name, *args, **args)

- To process jobs

mytaskman.run()

- To received finished data

def callback(data):
    # do something with data

mytaskman.run_finished(callback)

Once captured, it will delete the message from the queue

"""

import datetime
import time
import logging
from pickle import dumps, loads
from boto import sqs

QUEUED_POOL = "queued"
FINISHED_POOL = "finished"

class TaskMan(object):

    def __init__(self,
                 name,
                 aws_access_key_id,
                 aws_secret_access_key,
                 region='us-east-1',
                 ttl=7200
                 ):
        """
        :param name:
        :param aws_access_key_id:
        :param aws_secret_access_key:
        :param region:
        :param ttl:
        :return:
        """
        self._pools = [QUEUED_POOL, FINISHED_POOL]
        self.name = name
        self.ttl = ttl
        self.sqs_conn = sqs.connect_to_region(region,
                                       aws_access_key_id=aws_access_key_id,
                                       aws_secret_access_key=aws_secret_access_key)

    def queue(self, pool_key=QUEUED_POOL):
        name = self.name + "-" + pool_key
        queue = self.sqs_conn.get_queue(name)
        if not queue:
            queue = self.sqs_conn.create_queue(name)
        return queue

    def add(self, callable, delay_seconds=None, *args, **kwargs):
        """
        To add executable tasks
        :param callable:
        :param delay_seconds: Change the visibility delay for this message
        :param args:
        :param kwargs:
        :return:
        """
        m = self.write_message(QUEUED_POOL,
                               delay_seconds=delay_seconds,
                               data={
                                    'callable': callable,
                                    'args': args,
                                    'kwargs': kwargs
                               })
        return m.id

    def write_message(self, pool_key, data, delay_seconds=None):
        """
        Write the message to the pool
        :param pool_key: The pool key
        :param data: the data to save
        :param delay_seconds: Change the visibility delay for this message
        :return:
        """
        message = sqs.message.Message(body=dumps(data))
        m = self.queue(pool_key).write(message, delay_seconds=delay_seconds)
        return m

    def fetch(self, pool_key, size=5, wait_time=30):
        """
        Fetch the messages
        :param pool_key:
        :param size:
        :param wait_time:
        :return:
        """
        q = self.queue(pool_key)
        while q.count():
            for message in q.get_messages(num_messages=size,
                                          wait_time_seconds=wait_time):
                yield message

    def run(self, size=5, pause=5, burst=False, wait_time=20):
        """
        Run the task worker
        :param size: Total of tasks to request at a time
        :param pause: The pause time between each request
        :param burst: If true, it will run once
        :param wait_time: wait time between
        :return:
        """
        while True:
            for message in self.fetch(pool_key=QUEUED_POOL,
                                      size=size,
                                      wait_time=wait_time):
                try:
                    body = self.message_data_to_object(message)
                    result = self._run_task_callable(callable=body["callable"],
                                                     args=body["args"],
                                                     kwargs=body["kwargs"])
                    data = {
                        "id": message.id,
                        "status": "completed",
                        "result": result,
                        "created_at": datetime.datetime.now()
                    }
                except Exception as ex:
                    data = {
                        "id": message.id,
                        "status": "failed",
                        "message": ex.message,
                        "created_at": datetime.datetime.now()
                    }

                    logging.error("Failed running task: '%s'" % ex.message)
                finally:
                    self.write_message(FINISHED_POOL, data)
                    message.delete()
            if burst:
                break
            time.sleep(pause)

    def run_finished(self, callback, size=5, pause=5, burst=False,
                     wait_time=30, delete=True):
        """
        RUn finished queued and pass the message callable function
        :param callback: fun : callable function
        :param size:
        :param pause:
        :param burst: if true it will run one
        :param wait_time:
        :param delete: (bool) to delete upon processing
        :return:
        """
        while True:
            for message in self.fetch(pool_key=FINISHED_POOL,
                                      size=size,
                                      wait_time=wait_time):
                content = self.message_data_to_object(message)
                callback(content)
                if delete:
                    message.delete()

            if burst:
                break
            time.sleep(pause)

    def _run_task_callable(self, callable, args, kwargs):
        return callable(*args, **kwargs)

    def message_data_to_object(self, message):
        return loads(message.get_body())

    def delete(self):
        """
        Delete the Q and queue pools
        """
        for key in self._pools:
            self.sqs_conn.delete_queue(self.queue(key))





"""
To log error or events on RollBar

Add logging config in your flask_magic config

    LOGGING_CONFIG = {
        "version": 1,
        "handlers": {
            "rollbar": {
                "class": "magic_xxl.flask.rollbar_logger.RollbarLogger",
                "access_token": "",
                "level": "WARN"
            }
        },
        'loggers': {
            '': {
                'handlers': ['rollbar'],
                'level': 'WARN',
            },
        }
    }

"""
import os
from flask_magic import get_env, init_app
import rollbar
import rollbar.contrib.flask
from flask import got_request_exception, request
import logging
import logging.handlers

_env = get_env().lower()

class RollbarLogger(logging.Handler):
    def __init__(self, access_token,  *args, **kwargs):
        logging.Handler.__init__(self, *args, **kwargs)

        rollbar.init(

            # access token for the demo app: https://rollbar.com/demo
            access_token,

            # environment name
            _env,

            # server root directory, makes tracebacks prettier
            root=os.path.dirname(os.path.realpath(__file__)),

            # flask already sets up logging
            allow_logging_basic_config=False)

    def emit(self, record):
        if record.exc_info:
            rollbar.report_exc_info(record.exc_info)
        else:
            request = None

            rollbar.report_message(record.msg, record.levelname, request=request)

def rollbar_init_app(app):
    got_request_exception.connect(rollbar.contrib.flask.report_exception, app)

# Init app
init_app(rollbar_init_app)





"""
magic_xxl
"""

from setuptools import setup, find_packages
from __about__ import *

setup(
    name=__title__,
    version=__version__,
    license=__license__,
    author=__author__,
    author_email=__email__,
    description=__summary__,
    long_description=__doc__,
    url=__uri__,
    download_url='http://github.com/mardix/magic-xxl/tarball/master',
    py_modules=['magic_xxl'],
    keywords=["flask-magic",  "queue", "redis"],
    include_package_data=True,
    packages=find_packages(),
    platforms='any',
    install_requires=[
        'boto==2.39.0',
        'redis==2.10.5',
        "redis-collections==0.1.7",
        "rollbar==0.12.1",
        "newrelic==2.64.0.48",
        "mutagen==1.31",
        "elasticsearch-dsl==0.0.11",
        "elasticsearch==2.2.0"
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



"""
Audio

Requires
- SOX
- WAV2PNG
- WAV2JSON
- BPM-TOOLS

"""
from __future__ import division
import os
import subprocess
from tempfile import mkstemp
import requests
from mutagen.mp3 import MP3


def hex_color_to_rgba(value):
    """
    Convert HEX color to RGBA color - rrggbbaa (8chars)
    :param value:
    :return:
    """
    value = value.lower().lstrip('#')
    if len(value) == 8:
        return value
    if len(value) == 3:
        value = ''.join([v*2 for v in list(value)])
    return value + "ff"

def get_file_name(filename):
    """
    Return the filename without the path
    """
    return os.path.basename(filename)

def get_file_extension(filename):
    """
    Return a file extension, without the dot
    """
    return os.path.splitext(filename)[1][1:].lower()

def run(cmd):
    ps = subprocess.Popen(cmd,
                     shell=True,
                     stdout=subprocess.PIPE,
                     stderr=subprocess.STDOUT)
    output = ps.communicate()[0]
    return output


def get_from_url(url, dest=None):
    """
    Download from url
    :param url:
    :param dest:
    :return:
    """
    if not dest:
        _, dest = mkstemp()
        ext = url.split("?")[0] if "?" in url else url
        dest = "%s.%s" % (dest, get_file_extension(ext))
        os.close(_)

    with open(dest, 'wb') as handle:
        response = requests.get(url, stream=True)
        if not response.ok:
            raise ValueError("Unable to download file '%s'" % url)
        for block in response.iter_content(1024):
            handle.write(block)
    return dest

def is_valid(src):
    return True

def gen_waveform(src,
                 dest,
                 fgcolor="2e4562ff",
                 bgcolor="00000000",
                 channels="left right",
                 width=960,
                 height=100):
    """
    Generate the PNG and JSON waveform of an audio file

    :param src: the source file. ie: MP3
    :param dest: Destination where to place the file. ie: /my-path/xyz124/waveform
        it will return: /my-path/xyz124/waveform.json, /my-path/xyz124/waveform.png
    :param fgcolor: foreground color of the png. format: rrggbbaa hexadecimal
    :param bgcolor: background color of the png.  format: rrggbbaa hexadecimal
    :param channels: string of channels, default="left right"
    :param width: width of png
    :param height: height png
    :return: return a tuple (png_file, json_file) path
    """

    if not is_valid(src):
        raise ValueError("Invalid file format")

    bgcolor = hex_color_to_rgba(bgcolor or "00000000")
    fgcolor = hex_color_to_rgba(fgcolor or "2e4562ff")

    json_file = "%s.json" % dest
    png_file = "%s.png" % dest
    _, tmp_wav = mkstemp()
    os.close(_)


    sox_cmd = "sox {src} -c 2 -t wav {tmp_wav}"\
        .format(src=src, tmp_wav=tmp_wav)
    wav2json = "wav2json {tmp_wav} --channels {channels} -n -o {json_file}"\
        .format(tmp_wav=tmp_wav, json_file=json_file, channels=channels)
    wav2png = "wav2png {tmp_wav} " \
              "--foreground-color={fgcolor} --background-color={bgcolor} " \
              "--width {width} --height {height} " \
              "-o {png_file} "\
        .format(fgcolor=fgcolor,
                bgcolor=bgcolor,
                tmp_wav=tmp_wav,
                png_file=png_file,
                width=width,
                height=height)
    del_wav = "rm {tmp_wav}".format(tmp_wav=tmp_wav)

    cmd = sox_cmd + ";" + wav2png + ";" + wav2json + ";" + del_wav
    run(cmd)

    if not os.path.isfile(png_file) or not os.path.isfile(json_file):
        for f in [png_file, json_file]:
            if os.path.isfile(f):
                os.remove(f)
        raise IOError("Failed to save png or json file: %s - %s" % (png_file, json_file))

    return png_file, json_file

def get_bpm(src):
    cmd = "sox -G {src} -t raw -r 44100 -e float -c 1 - | bpm".format(src=src)
    r = run(cmd)
    r = r.replace("\n", "")
    return float(r)

def get_info(src):
    """
    Get the MP3 file info
    :param src:
    :return:
    """
    audio = MP3(src)
    info = audio.info
    bpm = get_bpm(src)
    return {
        "duration": round(info.length, 2),
        "channels": info.channels,
        "bitrate": info.bitrate // 1000,
        "bpm": int(round(bpm)),
        "filesize": round(os.path.getsize(src) / (1024**2), 2)
    }

def trim(src, dest):
    pass

def joint_stereo(src, dest):
    pass

def fadein(src, dest):
    pass

def fadeout(src, dest):
    pass